/*
 *  Copyright (C) 2015-2017  Thadeu Lima de Souza Cascardo <cascardo@minaslivre.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "calcula.h"
#include <errno.h>
#include <stdio.h>
#include "declaracao.h"
#include "cmd.h"
#include "rendimento.h"
#include "totais.h"
#include "util.h"
#include "ano.h"
#include "dependente.h"
#include "pagamento.h"

static const long long dependente[ANO(MAX_ANOS)] = {
	[ANO(2015)] = 215652,
	[ANO(2016)] = 227508,
	[ANO(2017)] = 227508,
};

static const long long instrucao[ANO(MAX_ANOS)] = {
	[ANO(2015)] = 337583,
	[ANO(2016)] = 356150,
	[ANO(2017)] = 356150,
};

long long deducao_dependente(struct declaracao *dec)
{
	if (ANO_VALIDO(dec->ano))
		return dependente[ANO(dec->ano)];
	return 0;
}

static void calculo_instrucao(struct declaracao *dec)
{
	int i, j;
	struct dependente *d;
	struct pagamento *p;
	long long instrucao_titular = 0;
	/* Instrução do titular */
	for (i = 0; (p = list_get(dec->pagamentos, i)); i++) {
		if (p->dependente == 0 && pagamento_instrucao(p)) {
			instrucao_titular += (p->pagamento - p->reembolso);
		}
	}
	if (instrucao_titular > instrucao[ANO(dec->ano)]) {
		totais_add(dec, "INSTRUCAO", instrucao[ANO(dec->ano)]);
	} else {
		totais_add(dec, "INSTRUCAO", instrucao_titular);
	}
	/* Dependentes com instrução */
	for (i = 0; (d = list_get(dec->dependentes, i)); i++) {
		long long instrucao_dependente = 0;
		for (j = 0; (p = list_get(dec->pagamentos, j)); j++) {
			if (p->dependente == (i + 1) && pagamento_instrucao(p)) {
				instrucao_dependente += (p->pagamento - p->reembolso);
			}
		}
		if (instrucao_dependente) {
			totais_add(dec, "DEPSINSTRUCAO", 1);
			if (dec->verbose) {
				printf("Dependente %s (%d) tem instrução\n", d->nome, i);
			}
		}
		if (instrucao_dependente > instrucao[ANO(dec->ano)]) {
			totais_add(dec, "INSTRUCAO", instrucao[ANO(dec->ano)]);
		} else {
			totais_add(dec, "INSTRUCAO", instrucao_dependente);
		}
	}
}

/* Alguns totais precisam ser limitados. Portanto, um total de decuções
 * precisa ser ajustado para tais limites. Esta função considerará tais
 * limites no futuro. */
static long long total_deducao(struct declaracao *dec)
{
	int i;

	calculo_instrucao(dec);

	if (dec->verbose) {
		printf("Dedução:\n");
		printf("\tDependentes: "FMT_R"\n", R(totais_get(dec, "DEPENDENTES")));
		printf("\tINSS: "FMT_R"\n", R(totais_get(dec, "INSS")));
		printf("\tInstrução: "FMT_R"\n", R(totais_get(dec, "INSTRUCAO")));
		printf("\tPagamentos: "FMT_R"\n", R(totais_get(dec, "PAGAMENTOS")));
		printf("\tReembolsos: -"FMT_R"\n", R(totais_get(dec, "REEMBOLSOS")));
	}
	return totais_get(dec, "DEPENDENTES") +
	       totais_get(dec, "INSS") +
	       totais_get(dec, "INSTRUCAO") +
	       totais_get(dec, "MEDICAS") +
	       totais_get(dec, "PREVIDENCIA");
}

static void total_pago(struct declaracao *dec)
{
	struct rendimento *rendimento;
	int i;
	dec->pago = dec->retido = 0;
	for (i = 0; rendimento = list_get(dec->rendimento, i); i++) {
		dec->pago += rendimento->imposto;
		dec->retido += rendimento->imposto;
	}
	if (dec->verbose) {
		printf("Total pago e retido: "FMT_R" "FMT_R"\n",
			R(dec->pago), R(dec->retido));
	}
}

struct taxtable {
	long long base;
	long long aliquota;
	long long deducao;
};

static struct taxtable table2015[] = {
	{       0,    0,      0, },
	{ 2145324,  750, 160899, },
	{ 3215148, 1500, 402035, },
	{ 4286917, 2250, 723554, },
	{ 5356572, 2750, 991383, },
	{ 9999999999999LL, 0, 0, },
};

static struct taxtable table2016[] = {
	{       0,    0,       0, },
	{ 2249914,  750,  168743, },
	{ 3347773, 1500,  419826, },
	{ 4447675, 2250,  753402, },
	{ 5537355, 2750, 1030270, },
	{ 9999999999999LL, 0, 0, },
};

static struct taxtable table2017[] = {
	{       0,    0,       0, },
	{ 2284777,  750,  171358, },
	{ 3391981, 1500,  425757, },
	{ 4501261, 2250,  763351, },
	{ 5597616, 2750, 1043232, },
	{ 9999999999999LL, 0, 0, },
};

static struct taxtable *table[ANO(MAX_ANOS)] = {
	[ANO(2015)] = table2015,
	[ANO(2016)] = table2016,
	[ANO(2017)] = table2017,
};

static const long long simples[ANO(MAX_ANOS)] = {
	[ANO(2015)] = 1588089,
	[ANO(2016)] = 1675434,
	[ANO(2017)] = 1675434,
};

static const long long obrigatoriedade[ANO(MAX_ANOS)] = {
	[ANO(2015)] = 2681655,
	[ANO(2016)] = 2812391,
	[ANO(2017)] = 2855970, /* De acordo com IN 1671/2016 */
};

static long long imposto(struct taxtable *tt, long long tr, int verbose)
{
	int i;
	for (i = 0; tr >= tt[i].base; i++);
	i--;
	if (verbose) {
		printf("Aplicando aliquota de %d%%, deduzindo " FMT_R"\n",
			tt[i].aliquota / 10, R(tt[i].deducao));
	}
	return tr * tt[i].aliquota / 10000 - tt[i].deducao;
}

static long long imposto_simples(struct declaracao *dec)
{
	struct taxtable *tt;
	long long tr, td;
	tt = table[ANO(dec->ano)];
	tr = totais_get(dec, "RENDTRIB");
	if (tr / 5 < simples[ANO(dec->ano)])
		td = tr / 5;
	else
		td = simples[ANO(dec->ano)];
	totais_add(dec, "DESCONTO", td);
	tr -= td;
	if (tr < 0)
		tr = 0;
	totais_add(dec, "BASESIMPLES", tr);
	if (dec->verbose) {
		printf("Desconto simplificado é "FMT_R"\n", R(td));
	}
	return imposto(tt, tr, dec->verbose);
}

static long long imposto_completa(struct declaracao *dec)
{
	struct taxtable *tt;
	long long tr, td;
	tt = table[ANO(dec->ano)];
	tr = totais_get(dec, "RENDTRIB");
	td = total_deducao(dec);
	totais_add(dec, "DEDUCOES", td);
	tr -= td;
	if (tr < 0)
		tr = 0;
	totais_add(dec, "BASECOMPLETA", tr);
	if (dec->verbose) {
		printf("Desconto completa é "FMT_R"\n", R(td));
	}
	return imposto(tt, tr, dec->verbose);
}

int calcula(struct declaracao *dec)
{
	long long i_simples, i_completa;
	if (!ANO_VALIDO(dec->ano)) {
		dec_set_error(dec, "Ano %d não suportado.", dec->ano);
		return -EINVAL;
	}
	if (totais_get(dec, "RENDTRIB") > obrigatoriedade[ANO(dec->ano)]) {
		if (dec->verbose) {
			printf("Declaracao obrigatoria pois rendimento e"
				"maior que mínimo para declaracao: "
				FMT_R" > "FMT_R"\n",
				R(totais_get(dec, "RENDTRIB")),
				R(obrigatoriedade[ANO(dec->ano)]));
		}
		dec->obrigatoria = 1;
	}
	i_simples = imposto_simples(dec);
	i_completa = imposto_completa(dec);
	total_pago(dec);
	if (dec->verbose) {
		printf("Imposto simplificada e completa: "FMT_R" "FMT_R"\n",
			R(i_simples), R(i_completa));
	}
	if (dec->tipo != FORCA_SIMPLES &&
	    (i_simples > i_completa || dec->tipo == FORCA_COMPLETA)) {
		totais_add(dec, "BASE", totais_get(dec, "BASECOMPLETA"));
		dec->tipo = COMPLETA;
		dec->devido = i_completa;
	} else {
		totais_add(dec, "BASE", totais_get(dec, "BASESIMPLES"));
		dec->tipo = SIMPLES;
		dec->devido = i_simples;
	}
	if (dec->pago > dec->devido)
		dec->restituicao = dec->pago - dec->devido;
	else
		dec->pagar = dec->devido - dec->pago;
	return 0;
}

static int run_calcula(struct declaracao *dec, char **args, int argc)
{
	totais_add(dec, "EXCLUSIVOS_SEM_13o",
		totais_get(dec, "EXCLUSIVOS") -
		totais_get(dec, "DECIMOTERCEIRO"));
	return calcula(dec);
}

static struct cmd cmd_calcula = {
	.name = "calcula",
	.run = run_calcula,
};

int calcula_cmd_init(void)
{
	cmd_add(&cmd_calcula);
	return 0;
}
